﻿using System;
using System.Collections.Generic;
using System.Text;

using System.IO;
using System.Net.Sockets;
using System.Text.RegularExpressions;
using System.Security.Cryptography;

namespace CKMail
{
    class CPop
    {
        private TcpClient _tcp;
        private NetworkStream _netstream;
        private StreamReader _reader;
        private StreamWriter _writer;

        /// <summary> [MailObjList]メソッド：CMailObjectのリストを返す
        /// </summary>
        /// <returns>Popオブジェクトのコレクション</returns>
        public List<CMailObject> MailObjList()
        {
            var lstMObject = new List<CMailObject>();

            // 接続を確認
            if (!_tcp.Connected)
            {
                CC.ShowMsg("接続されていません", CC.MSG_EXC);
                return lstMObject;
            }

            // ①LISTコマンドを実行して一覧を得る
            _writer.WriteLine("LIST");
            string result = _reader.ReadLine();
            if (!IsOK(result))
            {
                CC.ShowMsg("一覧取得に失敗しました:" + result, CC.MSG_EXC);
                return lstMObject;
            }

            // 「.」が出てくるまでループして一覧番号を得る
            var msglist = new Dictionary<int, int>();
            while ((result = _reader.ReadLine()) != ".")
            {
                string[] val = result.Split(new Char[] { ' ' });
                msglist.Add(int.Parse(val[0]), int.Parse(val[1]));
            }

            // メール数だけループ処理
            foreach (int key in msglist.Keys)
            {
                // ②RETRコマンドを実行してメール本体を得る
                _writer.WriteLine("RETR {0}", key);
                result = _reader.ReadLine();
                if (!IsOK(result))
                {
                    string msg1 = key.ToString()
                        + "番目のメール本体受信に失敗しました:" + result;
                    CC.ShowMsg(msg1, CC.MSG_EXC);
                    lstMObject.Clear();
                    return lstMObject;
                }

                // メール本文の読み取り
                // 「.」が出てくるまでループ
                bool isheader = true;
                var mobject = new CMailObject();
                string nowHeaderName = String.Empty;
                while ((result = _reader.ReadLine()) != ".")
                {
                    // 空行かどうか
                    if (result == String.Empty && isheader)
                    {
                        // 次からボディ
                        isheader = false;
                        continue;
                    }

                    if (result.StartsWith("."))
                    {
                        // 「.」をひとつトル
                        result = result.Substring(1);
                    }

                    if (isheader)
                    {
                        // 先頭がスペースやタブなら、前の行とつなげる
                        if (result.StartsWith(" ") || result.StartsWith("\t"))
                        {
                            mobject.Header[nowHeaderName] += result;
                        }
                        else
                        {
                            // 「:」区切りでヘッダの値を取得
                            string[] val = result.Split(new Char[] { ':' }, 2);
                            nowHeaderName = val[0].Trim();
                            mobject.Header.Add(nowHeaderName, val[1].Trim());
                        }
                    }
                }

                // Ver1.3.0からtry～catchでくくり、Encodeの不適切なメールをカット
                try   
                {
                    // デコードして整形する
                    mobject.ParseHeader();
                    lstMObject.Add(mobject);
                }
                catch
                { }

            }
            return lstMObject;
        }

        /// <summary> [IsConnectOK]メソッド：サーバーに接続して認証する
        /// </summary>
        /// <param name="vservername">サーバー名</param>
        /// <param name="vport">ポート番号</param>
        /// <param name="vusername">ユーザー名</param>
        /// <param name="vpassword">パスワード</param>
        /// <param name="vfromwatch">[F20_WatchMail]からの場合のみ true</param>
        /// <returns></returns>
        public bool IsConnectOK(string vservername,
                                int vport,
                                string vusername,
                                string vpassword,
                                bool vfromwatch)
        {
            // 接続を開く
            _tcp = new TcpClient();
            try
            {
                _tcp.Connect(vservername, vport);
            }
            catch
            {
                if (!vfromwatch)
                {
                    string msg1 = "メールサーバーとの接続ができません。"
                        + CC.NL + CC.NL
                        + "インターネットとの接続、またはPOP3サーバー名、"
                        + "ポートの設定に問題がある可能性があります。";
                    CC.ShowMsg(msg1, CC.MSG_EXC);
                }
                _tcp = null;
                return false;
            }

            // NetworkStreamを得る
            _netstream = _tcp.GetStream();
            // タイムアウト時間の設定（ここでは仮に30秒）
            _netstream.ReadTimeout = 30 * 1000;
            _netstream.WriteTimeout = 30 * 1000;

            // StreamReader、StreamWriterを被せる
            _reader = new StreamReader(_netstream, Encoding.ASCII);
            _writer = new StreamWriter(_netstream, Encoding.ASCII);

            // 改行文字とバッファフラッシュを設定
            _writer.NewLine = "\r\n";
            _writer.AutoFlush = true;
            // サーバーからの挨拶コードを得る
            string result = _reader.ReadLine();
            if (!IsOK(result))
            {
                // OKではない
                CC.ShowMsg("サーバーからの応答コードが不正です:" + result, CC.MSG_EXC);
                Quit();
                return false;
            }

            // パスワード認証、APOPの場合は「<」と「>」の中身が必要なので取り出す
            Match challenge = Regex.Match(result, "<([^<>]+)>");
            if (!challenge.Success)
            {
                CC.ShowMsg("APOP認証のための識別子が見つかりません:" + result, CC.MSG_EXC);
                Quit();
                return false;
            }
            if (IsApopOK(vusername, vpassword, challenge.Groups[1].ToString()))
            {
                return true;
            }
            else
            {
                Quit();
                return false;
            }
        }

        /// <summary> [Quit]メソッド：サーバーから切断する
        /// </summary>
        public void Quit()
        {
            // サーバーから切断する
            if (_tcp.Connected)
            {
                // QUITコマンドを送信して切断する
                _writer.WriteLine("QUIT");
                string result = _reader.ReadLine();
                if (!IsOK(result))
                    CC.ShowMsg("切断時にエラーが発生しました:" + result, CC.MSG_EXC);
                // 接続を閉じる
                _netstream.Dispose();
                _reader.Dispose();
                _writer.Dispose();
                _tcp.Close();
                _tcp = null;
            }
        }


        /// <summary> [IsApopAuth]メソッド：APOP認証をして、認証成否を返す
        /// [IsConnectOK]から呼出し
        /// </summary>
        /// <param name="vusername">ユーザー名</param>
        /// <param name="vpassword">パスワード</param>
        /// <param name="vchallenge">
        /// POP3サーバーの応答の1行目から得られるチャレンジ文字列
        /// </param>
        /// <returns></returns>
        private bool IsApopOK(string vusername, string vpassword, string vchallenge)
        {
            // パスワードを結合
            string plain = "<" + vchallenge + ">" + vpassword;

            // MD5に変換
            MD5 md5 = MD5.Create();
            byte[] data = md5.ComputeHash(Encoding.ASCII.GetBytes(plain));

            // 16進数小文字の文字列にする
            StringBuilder md5response = new StringBuilder();
            for (int i = 0; i < data.Length; i++)
            {
                md5response.Append(data[i].ToString("x2"));
            }

            // APOPコマンド送信
            _writer.WriteLine("APOP {0} {1}", vusername, md5response.ToString());
            string result = _reader.ReadLine();
            if (IsOK(result))
            {
                return true;
            }
            else
            {
                // OKではない
                string msg = "認証に失敗しました:" + result + CC.NL
                    + "アカウント名、パスワードの設定に問題があるかもしれません。";
                CC.ShowMsg(msg, CC.MSG_EXC);
                return false;
            }
        }

        /// <summary> [IsOK]メソッド：応答行が「+OK」か「-ERR」かを調べる
        /// </summary>
        /// <param name="vresult"></param>
        /// <returns></returns>
        private bool IsOK(string vrow)
        {
            if (vrow.StartsWith("+OK"))
                return true;
            else
                return false;
        }

    }
}
